# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Several basic tests for hierarchical clustering procedures.

Should be cast soon in a nicer unitest framework

Author : Bertrand Thirion, 2008-2009
"""

import numpy as np
from numpy.random import randn
from ..hierarchical_clustering import (average_link_graph,
                                       average_link_graph_segment,
                                       ward, ward_quick,
                                       ward_segment, ward_field_segment,
                                       ward_quick_segment)
from nipy.algorithms.graph.graph import WeightedGraph, knn
from nipy.algorithms.graph.field import field_from_graph_and_data

from nose.tools import assert_true, assert_equal


def alg_test_basic(n=100,k=5):
    # Check that we obtain the correct solution in a simplistic case
    np.random.seed(0)
    x = np.random.randn(n, 2)
    x[:int(0.7*n)] += 3
    G = knn(x, k)
    t = average_link_graph(G)
    u = t.split(2)
    v = np.zeros(n)
    v[:int(0.7*n)]=1
    w = np.absolute(u-v)
    assert_true(np.sum(w*(1-w))==0)


def alg_test_2():
    # Do we handle case of graph with too many connected components?
    np.random.seed(0)
    n = 100
    k = 5
    x = np.random.randn(n, 2)
    x[:int(0.3*n)] += 10
    x[int(0.8*n):] -= 10
    G = knn(x, k)
    t = average_link_graph(G)
    u = t.split(2)
    assert_true(u.max()==2)


def alg_test_3(n=100,k=5):
    # Check that we obtain the correct solution in a simplistic case
    np.random.seed(0)
    x = np.random.randn(n, 2)
    x[:int(0.7*n)] += 3
    G = knn(x, k)
    u, cost = average_link_graph_segment(G, qmax=2)
    v = np.zeros(n)
    v[:int(0.7*n)]=1
    w = np.absolute(u-v)
    assert_true(np.sum(w*(1-w))==0)


def ward_test_basic(n=100,k=5):
    # Basic check of ward's algorithm
    np.random.seed(0)
    x = np.random.randn(n, 2)
    x[:int(0.7*n)] += 3
    G = knn(x, k)
    t = ward(G,x)
    u = t.split(2)
    v = np.zeros(n)
    v[:int(0.7*n)]=1
    w = np.absolute(u-v)
    assert_true(np.sum(w*(1-w))==0)


def wardq_test_basic(n=100,k=5):
    # Basic check of ward's algorithm
    np.random.seed(0)
    x = np.random.randn(n, 2)
    x[:int(0.7*n)] += 3
    G = knn(x, k)
    t = ward_quick(G, x)
    u = t.split(2)
    v = np.zeros(n)
    v[:int(0.7*n)]=1
    w = np.absolute(u-v)
    assert_true(np.sum(w*(1-w))==0)


def wardq_test_2():
    # Do we handle case of graph with too many connected components?
    np.random.seed(0)
    n = 100
    k = 5
    x = np.random.randn(n, 2)
    x[:int(0.3*n)] += 10
    x[int(0.8*n):] -= 10
    G = knn(x, k)
    t = ward_quick(G, x)
    u = t.split(2)
    assert_equal(u.max(), 2)


def wardf_test(n=100,k=5):
    np.random.seed(0)
    x = np.random.randn(n,2)
    x[:int(0.7*n)] += 3
    G = knn(x, 5)
    F = field_from_graph_and_data(G, x)
    u, cost = ward_field_segment(F, qmax=2)
    v = np.zeros(n)
    v[:int(0.7*n)]=1
    w = np.absolute(u-v)
    assert_equal(np.sum(w*(1-w)), 0)


def wards_test_basic(n=100,k=5):
    # Basic check of ward's segmentation algorithm
    np.random.seed(0)
    x = np.random.randn(n, 2)
    x[:int(0.7*n)] += 3
    G = knn(x, k)
    u,cost =  ward_segment(G, x, qmax=2)
    v = np.zeros(n)
    v[:int(0.7*n)]=1
    w = np.absolute(u-v)
    assert_equal(np.sum(w*(1-w)), 0)


def wards_test_3():
    # Check ward_segment
    np.random.seed(0)
    n = 100
    k = 5
    x = np.random.randn(n,2)
    x[:int(0.3*n)] += 10
    x[int(0.8*n):] -= 10
    G = knn(x,k)
    u,cost = ward_segment(G, x, qmax=2)
    assert_equal(u.max(), 2)


def cost_test(n=100, k=5):
    # check that cost.max() is equal to the data variance
    np.random.seed(0)
    x = np.random.randn(n, 2)
    G = knn(x, k)
    u, cost =  ward_segment(G, x)
    assert_true(np.abs(cost.max()/(n*np.var(x,0).sum()) - 1) < 1e-6)


def ward_test_more(n=100, k=5, verbose=0):
    # Check that two implementations give the same result
    np.random.seed(0)
    X = randn(n,2)
    X[:np.ceil(n/3)] += 5
    G = knn(X, 5)
    u,c = ward_segment(G, X, stop=-1, qmax=1, verbose=verbose)
    u1,c = ward_segment(G, X, stop=-1, qmax=k, verbose=verbose)
    u,c = ward_quick_segment(G, X, stop=-1, qmax=1, verbose=verbose)
    u2,c = ward_quick_segment(G, X, stop=-1, qmax=k, verbose=verbose)
    assert_equal(np.sum(u1==u2), n)


if __name__ == '__main__':
    import nose
    nose.run(argv=['', __file__])



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This test basically just plays around with image.rollimg.

It has three examples

* image_reduce:  this takes an Image having, say, an axis 't' and returns another
  Image having reduced over 't'

* need_specific_axis_reduce: this takes an Image and a specific
  axis name, like 't' and produces an Image reduced over 't'. raises an
  exception if Image has no axis 't'

* image_call: this takes an Image having, say, an axis 't'
  and does something along this axis -- like fits a regression model? and
  outputs a new Image with the 't' axis replaced by 'new'

* image_modify_copy: this takes an Image and an axis specification,
  such as 'x+LR', 'l', or 2, modifies a copy of the data by iterating over this
  axis, and returns an Image with the same axes


Notes
-----

In these loaded Images, 't' is both an axis name and a world coordinate name so
it is not ambiguous to say 't' axis. It is slightly ambiguous to say 'x+LR' axis
if the axisnames are ['slice', 'frequency', 'phase'] but image.rollimg
identifies 'x+LR' == 'slice' == 0.
"""

import numpy as np

from ..image import (Image, rollimg, synchronized_order)
from ...reference.coordinate_map import (AffineTransform as AT, drop_io_dim,
                                         AxisError)
from ...reference.coordinate_system import CoordinateSystem as CS
from ...reference.spaces import mni_csm
from ...image.image_spaces import xyz_affine

from nose.tools import (assert_raises, assert_equal)

from numpy.testing import assert_almost_equal, assert_array_equal


MNI3 = mni_csm(3)
MNI4 = mni_csm(4)


def image_reduce(img, reduce_op, axis='t'):
    """
    Take an Image, perform some reduce operation on it, over
    a specified axis, and return a new Image.

    For the sake of testing things out, we will assume that
    the operation reduces over the first axis only.

    Parameters
    ----------
    image : Image
    reduce_op : callable
        An operation that reduces over the first axis,
        such as lambda x: x.sum(0)
    axis : str or int
        Specification of axis of Image

    Returns
    -------
    newim : Image, missing axis
    """
    img = rollimg(img, axis)
    axis_name = img.axes.coord_names[0]
    output_axes = list(img.axes.coord_names)
    output_axes.remove(axis_name)
    newdata = reduce_op(img.get_data())
    return Image(newdata, drop_io_dim(img.coordmap, axis))


def need_specific_axis_reduce(img, reduce_op):
    """
    Take an Image, perform some reduce operation on it, over the axis named
    'specific', and return a new Image.

    For the sake of testing things out, we will assume that the operation
    reduces over the first axis only.

    Parameters
    ----------
    img : Image
    reduce_op : callable
        An operation that reduces over the first axis,
        such as lambda x: x.sum(0)

    Returns
    -------
    newim : Image, missing axis
    """
    return image_reduce(img, reduce_op, 'specific')


def image_call(img, function, inaxis='t', outaxis='new'):
    """
    Take an Image, perform some operation on it, over a specified axis, and
    return a new Image.

    For the sake of testing things out, we will assume that the operation can
    only operate on the first axis of the array.

    Parameters
    ----------
    img : Image
    function : callable
        An operation that does something over the first axis,
        such as lambda x: x[::2]
    inaxis : str or int
        Specification of axis of Image
    outaxis : str
        Name of new axis in new Image

    Returns
    -------
    newim : Image
        with axis `inaxis` replaced with `outaxis`
    """
    rolled_img = rollimg(img, inaxis)
    inaxis = rolled_img.axes.coord_names[0] # now it's a string
    newdata = function(rolled_img.get_data())
    new_coordmap = rolled_img.coordmap.renamed_domain({inaxis: outaxis})
    new_image = Image(newdata, new_coordmap)
    # we have to roll the axis back
    axis_index = img.axes.index(inaxis) + 1
    return rollimg(new_image, 0, axis_index)


def image_modify(img, modify, axis='y+PA'):
    """
    Take an Image, perform some operation on it, over a specified axis, and
    return a new Image.

    For this operation, we are allowed to iterate over spatial axes.

    For the sake of testing things out, we will assume that the operation modify
    can only operate by iterating over the first axis of an array.

    Parameters
    ----------
    img : Image
    modify : callable
        An operation that modifies an array.  Something like::

            def f(x):
                x[:] = x.mean()

    axis : str or int
        Specification of axis of Image

    Returns
    -------
    newim : Image
        with a modified copy of img._data.
    """
    rolled_img = rollimg(img, axis)
    data = rolled_img.get_data().copy()
    for d in data:
        modify(d)
    import copy
    new_image = Image(data, copy.copy(rolled_img.coordmap))
    # Now, we have to put the data back to same order as img
    return synchronized_order(new_image, img)


def test_reduce():
    shape = (3, 5, 7, 9)
    x = np.random.standard_normal(shape)
    im = Image(x, AT(CS('ijkq'), MNI4, np.diag([3, 4, 5, 6, 1])))
    newim = image_reduce(im, lambda x: x.sum(0), 'q')
    assert_array_equal(xyz_affine(im), xyz_affine(newim))
    assert_equal(newim.axes.coord_names, tuple('ijk'))
    assert_equal(newim.shape, (3, 5, 7))
    assert_almost_equal(newim.get_data(), x.sum(3))
    im_nd = Image(x, AT(CS('ijkq'), MNI4, np.array(
        [[0, 1, 2, 0, 10],
         [3, 4, 5, 0, 11],
         [6, 7, 8, 0, 12],
         [0, 0, 0, 9, 13],
         [0, 0, 0, 0, 1]])))
    for i, o, n in zip('ijk', MNI3.coord_names, range(3)):
        for axis_id in (i, o, n):
            # Non-diagonal reduce raise an error
            assert_raises(AxisError, image_reduce, im_nd,
                          lambda x: x.sum(0), axis_id)
            # Diagonal reduces are OK
            newim = image_reduce(im, lambda x: x.sum(0), axis_id)


def test_specific_reduce():
    shape = (3, 5, 7, 9)
    x = np.random.standard_normal(shape)
    im = Image(x, AT(CS('ijkq'), MNI4, np.diag([3, 4, 5, 6, 1])))
    # we have to rename the axis before we can call the function
    # need_specific_axis_reduce on it
    assert_raises(AxisError, need_specific_axis_reduce, im, lambda x: x.sum(0))
    im = im.renamed_axes(q='specific')
    newim = need_specific_axis_reduce(im, lambda x: x.sum(0))
    assert_array_equal(xyz_affine(im), xyz_affine(newim))
    assert_equal(newim.axes.coord_names, tuple('ijk'))
    assert_equal(newim.shape, (3, 5, 7))
    assert_almost_equal(newim.get_data(), x.sum(3))


def test_call():
    shape = (3, 5, 7, 12)
    x = np.random.standard_normal(shape)
    affine = np.eye(5)
    affine[:3, :3] = np.random.standard_normal((3, 3))
    affine[:4, 4] = np.random.standard_normal((4,))
    im = Image(x, AT(CS('ijkq'), MNI4, affine))
    newim = image_call(im, lambda x: x[::2], 'q', 'out')
    assert_array_equal(xyz_affine(im), xyz_affine(newim))
    assert_equal(newim.axes.coord_names, tuple('ijk') + ('out',))
    assert_equal(newim.shape, (3, 5, 7, 6))
    assert_almost_equal(newim.get_data(), x[:,:,:,::2])


def test_modify():
    shape = (3, 5, 7, 12)
    x = np.random.standard_normal(shape)
    affine = np.eye(5)
    affine[:3, :3] = np.random.standard_normal((3, 3))
    affine[:4, 4] = np.random.standard_normal((4,))
    im = Image(x, AT(CS('ijkq'), MNI4, affine))

    def nullmodify(d):
        pass

    def meanmodify(d):
        d[:] = d.mean()

    for i, o, n in zip('ijkq', MNI3.coord_names + ('q',), range(4)):
        for a in i, o, n:
            nullim = image_modify(im, nullmodify, a)
            meanim = image_modify(im, meanmodify, a)
            assert_array_equal(nullim.get_data(), im.get_data())
            assert_array_equal(xyz_affine(im), xyz_affine(nullim))
            assert_equal(nullim.axes, im.axes)
            # yield assert_equal, nullim, im
            assert_array_equal(xyz_affine(im), xyz_affine(meanim))
            assert_equal(meanim.axes, im.axes)
        # Make sure that meanmodify works as expected
        d = im.get_data()
        d = np.rollaxis(d, n)
        meand = meanim.get_data()
        meand = np.rollaxis(meand, n)
        for i in range(d.shape[0]):
            assert_almost_equal(meand[i], d[i].mean())



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Use Mayavi to visualize the structure of a VolumeData
"""

from enthought.mayavi import mlab
import numpy as np

x, y, z, s = np.random.random((4, 20))

mlab.figure(1, fgcolor=(0, 0, 0), bgcolor=(1, 1, 1))
mlab.clf()

src = mlab.pipeline.scalar_scatter(x, y, z, s)
sgrid = mlab.pipeline.delaunay3d(src)

mlab.pipeline.surface(sgrid, opacity=0.4)
mlab.pipeline.surface(mlab.pipeline.extract_edges(sgrid), color=(0, 0, 0))
mlab.pipeline.glyph(sgrid, mode='cube', scale_factor=0.05, scale_mode='none')
mlab.savefig('volume_data.jpg')
mlab.show()






#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
__doc__ = """
This shows the effect of ward clustering on a real fMRI dataset

Author: Bertrand Thirion, 2010
"""
print(__doc__)

from os import mkdir, getcwd, path

import numpy as np

from nibabel import load, save, Nifti1Image

from nipy.algorithms.graph.field import Field

# Local import
from get_data_light import DATA_DIR, get_second_level_dataset

# paths
input_image = path.join(DATA_DIR, 'spmT_0029.nii.gz')
mask_image = path.join(DATA_DIR, 'mask.nii.gz')
if (not path.exists(mask_image)) or (not path.exists(input_image)):
    get_second_level_dataset()

# write directory
write_dir = path.join(getcwd(), 'results')
if not path.exists(write_dir):
    mkdir(write_dir)

# read the data
mask = load(mask_image).get_data() > 0
ijk = np.array(np.where(mask)).T
nvox = ijk.shape[0]
data = load(input_image).get_data()[mask]
image_field = Field(nvox)
image_field.from_3d_grid(ijk, k=6)
image_field.set_field(data)
u, _ = image_field.ward(100)

# write the results
label_image = path.join(write_dir, 'label.nii')
wdata = mask - 1
wdata[mask] = u
save(Nifti1Image(wdata, load(mask_image).get_affine()), label_image)
print("Label image written in %s" % label_image)



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
__doc__ = """
Examples of design matrices specification and and computation (event-related
design, FIR design, etc)

Requires matplotlib

Author : Bertrand Thirion: 2009-2010
"""
print(__doc__)

import numpy as np

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError("This script needs the matplotlib library")

from nipy.modalities.fmri.design_matrix import make_dmtx
from nipy.modalities.fmri.experimental_paradigm import (EventRelatedParadigm,
                                                        BlockParadigm)

# frame times
tr = 1.0
nscans = 128
frametimes = np.linspace(0, (nscans - 1) * tr, nscans)

# experimental paradigm
conditions = ['c0', 'c0', 'c0', 'c1', 'c1', 'c1', 'c3', 'c3', 'c3']
onsets = [30, 70, 100, 10, 30, 90, 30, 40, 60]
hrf_model = 'canonical'
motion = np.cumsum(np.random.randn(128, 6), 0)
add_reg_names = ['tx', 'ty', 'tz', 'rx', 'ry', 'rz']

#event-related design matrix
paradigm = EventRelatedParadigm(conditions, onsets)

X1 = make_dmtx(
    frametimes, paradigm, drift_model='polynomial', drift_order=3,
    add_regs=motion, add_reg_names=add_reg_names)

# block design matrix
duration = 7 * np.ones(9)
paradigm = BlockParadigm(con_id=conditions, onset=onsets,
                             duration=duration)

X2 = make_dmtx(frametimes, paradigm, drift_model='polynomial',
               drift_order=3)

# FIR model
paradigm = EventRelatedParadigm(conditions, onsets)
hrf_model = 'FIR'
X3 = make_dmtx(frametimes, paradigm, hrf_model='fir',
               drift_model='polynomial', drift_order=3,
               fir_delays=np.arange(1, 6))

# plot the results
fig = plt.figure(figsize=(10, 6))
ax = plt.subplot(1, 3, 1)
X1.show(ax=ax)
ax.set_title('Event-related design matrix', fontsize=12)
ax = plt.subplot(1, 3, 2)
X2.show(ax=ax)
ax.set_title('Block design matrix', fontsize=12)
ax = plt.subplot(1, 3, 3)
X3.show(ax=ax)
ax.set_title('FIR design matrix', fontsize=12)
plt.subplots_adjust(top=0.9, bottom=0.25)
plt.show()



"""
Test the Infinite GMM.

Author : Bertrand Thirion, 2010
"""

import numpy as np

from ..imm import IMM, MixedIMM, co_labelling

from nose.tools import assert_true

from numpy.testing import assert_array_equal

def test_colabel():
    # test the co_labelling functionality
    z = np.array([0,1,1,0,2])
    c = co_labelling(z).todense()
    tc = np.array([[ 1.,  0.,  0.,  1.,  0.],
                   [ 0.,  1.,  1.,  0.,  0.],
                   [ 0.,  1.,  1.,  0.,  0.],
                   [ 1.,  0.,  0.,  1.,  0.],
                   [ 0.,  0.,  0.,  0.,  1.]])
    assert_array_equal(c, tc)


def test_imm_loglike_1D():
    # Check that the log-likelihood of the data under the infinite gaussian
    # mixture model is close to the theoretical data likelihood
    n = 100
    dim = 1
    alpha = .5
    x = np.random.randn(n, dim)
    igmm = IMM(alpha, dim)
    igmm.set_priors(x)

    # warming
    igmm.sample(x, niter=100)

    # sampling
    like =  igmm.sample(x, niter=300)
    theoretical_ll = -dim*.5*(1+np.log(2*np.pi))
    empirical_ll = np.log(like).mean()
    assert_true(np.absolute(theoretical_ll-empirical_ll)<0.25*dim)


def test_imm_loglike_known_groups():
    # Check that the log-likelihood of the data under IGMM close to theory
    n = 50
    dim = 1
    alpha = .5
    x = np.random.randn(n, dim)
    igmm = IMM(alpha, dim)
    igmm.set_priors(x)
    kfold = np.floor(np.random.rand(n)*5).astype(np.int)

    # warming
    igmm.sample(x, niter=100)

    # sampling
    like =  igmm.sample(x, niter=300, kfold=kfold)
    theoretical_ll = -dim*.5*(1+np.log(2*np.pi))
    empirical_ll = np.log(like).mean()
    assert_true(np.absolute(theoretical_ll-empirical_ll)<0.25*dim)


def test_imm_loglike_1D_k10():
    # Check with k-fold cross validation (k=10)
    n = 50
    dim = 1
    alpha = .5
    k = 5
    x = np.random.randn(n, dim)
    igmm = IMM(alpha, dim)
    igmm.set_priors(x)

    # warming
    igmm.sample(x, niter=100, kfold=k)

    # sampling
    like =  igmm.sample(x, niter=300, kfold=k)
    theoretical_ll = -dim*.5*(1+np.log(2*np.pi))
    empirical_ll = np.log(like).mean()
    assert_true(np.absolute(theoretical_ll-empirical_ll)<0.25*dim)


def test_imm_loglike_2D_fast():
    # Faster version for log-likelihood imm
    n = 100
    dim = 2
    alpha = .5
    x = np.random.randn(n, dim)
    igmm = IMM(alpha, dim)
    igmm.set_priors(x)

    # warming
    igmm.sample(x, niter=100, init=True)

    # sampling
    like =  igmm.sample(x, niter=300)
    theoretical_ll = -dim*.5*(1+np.log(2*np.pi))
    empirical_ll = np.log(like).mean()
    assert_true(np.absolute(theoretical_ll-empirical_ll)<0.25*dim)


def test_imm_loglike_2D():
    # Slower cross-validated logL check
    n = 50
    dim = 2
    alpha = .5
    k = 5
    x = np.random.randn(n, dim)
    igmm = IMM(alpha, dim)
    igmm.set_priors(x)

    # warming
    igmm.sample(x, niter=100, init=True, kfold=k)

    # sampling
    like =  igmm.sample(x, niter=300, kfold=k)
    theoretical_ll = -dim*.5*(1+np.log(2*np.pi))
    empirical_ll = np.log(like).mean()
    assert_true(np.absolute(theoretical_ll-empirical_ll)<0.25*dim)


def test_imm_loglike_2D_a0_1():
    # Check with alpha=.1
    n = 100
    dim = 2
    alpha = .1
    x = np.random.randn(n, dim)
    igmm = IMM(alpha, dim)
    igmm.set_priors(x)

    # warming
    igmm.sample(x, niter=100, init=True)

    # sampling
    like =  igmm.sample(x, niter=300)
    theoretical_ll = -dim*.5*(1+np.log(2*np.pi))
    empirical_ll = np.log(like).mean()
    print theoretical_ll, empirical_ll
    assert_true(np.absolute(theoretical_ll-empirical_ll)<0.2*dim)


def test_imm_wnc():
    # Test the basic imm_wnc
    n = 50
    dim = 1
    alpha = .5
    g0 = 1.
    x = np.random.rand(n, dim)
    x[:.3*n] *= .2
    x[:.1*n] *= .3

    # instantiate
    migmm = MixedIMM(alpha, dim)
    migmm.set_priors(x)
    migmm.set_constant_densities(null_dens=g0)
    ncp = 0.5*np.ones(n)

    # warming
    migmm.sample(x, null_class_proba=ncp, niter=100, init=True)
    g = np.reshape(np.linspace(0, 1, 101), (101, dim))

    # sampling
    like, pproba =  migmm.sample(x, null_class_proba=ncp, niter=300,
                                 sampling_points=g)

    # the density should sum to 1
    ds = 0.01*like.sum()
    assert_true(ds<1)
    assert_true(ds>.8)
    assert_true(np.sum(pproba>.5)>1)
    assert_true(np.sum(pproba<.5)>1)


def test_imm_wnc1():
    # Test the basic imm_wnc, where the probaility under the null is random
    n = 50
    dim = 1
    alpha = .5
    g0 = 1.
    x = np.random.rand(n, dim)
    x[:.3*n] *= .2
    x[:.1*n] *= .3

    # instantiate
    migmm = MixedIMM(alpha, dim)
    migmm.set_priors(x)
    migmm.set_constant_densities(null_dens=g0)
    ncp = np.random.rand(n)

    # warming
    migmm.sample(x, null_class_proba=ncp, niter=100, init=True)
    g = np.reshape(np.linspace(0, 1, 101), (101, dim))

    #sampling
    like, pproba =  migmm.sample(x, null_class_proba=ncp, niter=300,
                         sampling_points=g)

    # the density should sum to 1
    ds = 0.01*like.sum()
    assert_true(ds<1)
    assert_true(ds>.8)
    assert_true(np.sum(pproba>.5)>1)
    assert_true(np.sum(pproba<.5)>1)


def test_imm_wnc2():
    # Test the basic imm_wnc when null class is shrunk to 0
    n = 50
    dim = 1
    alpha = .5
    g0 = 1.
    x = np.random.rand(n, dim)
    x[:.3*n] *= .2
    x[:.1*n] *= .3

    # instantiate
    migmm = MixedIMM(alpha, dim)
    migmm.set_priors(x)
    migmm.set_constant_densities(null_dens=g0)
    ncp = np.zeros(n)

    # warming
    migmm.sample(x, null_class_proba=ncp, niter=100, init=True)

    # sampling
    like, pproba =  migmm.sample(x, null_class_proba=ncp, niter=300)
    assert_true(like.min()>.1)
    assert_true(like.max()<5.)
    assert_array_equal(pproba, ncp)


def test_imm_wnc3():
    # Test the basic imm_wnc when null class is of prob 1 (nothing is estimated)
    n = 50
    dim = 1
    alpha = .5
    g0 = 1.
    x = np.random.rand(n, dim)
    x[:.3*n] *= .2
    x[:.1*n] *= .3

    # instantiate
    migmm = MixedIMM(alpha, dim)
    migmm.set_priors(x)
    migmm.set_constant_densities(null_dens=g0)
    ncp = np.ones(n)

    # warming
    migmm.sample(x, null_class_proba=ncp, niter=100, init=True)

    # sampling
    like, pproba =  migmm.sample(x, null_class_proba=ncp, niter=300)
    assert_array_equal(pproba, ncp)


if __name__ == '__main__':
    import nose
    nose.run(argv=['', __file__])



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import time

import numpy as np

from .. import copy_vector


def time_ratio(t0,t1):
    if t1==0:
        return np.inf
    else:
        return t0/t1

def time_copy_vector(x):
    t0 = time.clock()
    y0 = copy_vector(x, 0) 
    dt0 = time.clock()-t0
    t1 = time.clock()
    y1 = copy_vector(x, 1) 
    dt1 = time.clock()-t1
    ratio = time_ratio(dt0,dt1)
    print('  using fff_array: %f sec' % dt0)
    print('  using numpy C API: %f sec' % dt1)
    print('  ratio: %f' % ratio)

def bench_copy_vector_contiguous(): 
    x = (1000*np.random.rand(1e6)).astype('int32')
    print('Contiguous buffer copy (int32-->double)')
    time_copy_vector(x)

def bench_copy_vector_strided(): 
    x0 = (1000*np.random.rand(2e6)).astype('int32')
    x = x0[::2]
    print('Non-contiguous buffer copy (int32-->double)')
    time_copy_vector(x)




#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
from os.path import join as pjoin, exists
import sys
from glob import glob
from distutils import log

# BEFORE importing distutils, remove MANIFEST. distutils doesn't properly
# update it when the contents of directories change.
if exists('MANIFEST'): os.remove('MANIFEST')

# Import build helpers
try:
    from nisext.sexts import package_check, get_comrec_build
except ImportError:
    raise RuntimeError('Need nisext package from nibabel installation'
                       ' - please install nibabel first')

from setup_helpers import (generate_a_pyrex_source,
                           cmdclass, INFO_VARS,
                           build_py) # build_py will do 2to3 for Python3

# monkey-patch numpy distutils to use Cython instead of Pyrex
from numpy.distutils.command.build_src import build_src
build_src.generate_a_pyrex_source = generate_a_pyrex_source

# Add custom commit-recording build command
cmdclass['build_py'] = get_comrec_build('nipy', build_py)

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration(None, parent_package, top_path)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)
    # The quiet=True option will silence all of the name setting warnings:
    # Ignoring attempt to set 'name' (from 'nipy.core' to 
    #    'nipy.core.image')
    # Robert Kern recommends setting quiet=True on the numpy list, stating
    # these messages are probably only used in debugging numpy distutils.
    config.get_version(pjoin('nipy', 'info.py')) # sets config.version
    config.add_subpackage('nipy', 'nipy')
    return config

################################################################################
# For some commands, use setuptools

if len(set(('develop', 'bdist_egg', 'bdist_rpm', 'bdist', 'bdist_dumb',
            'install_egg_info', 'egg_info', 'easy_install', 'bdist_mpkg',
            )).intersection(sys.argv)) > 0:
    from setup_egg import extra_setuptools_args

# extra_setuptools_args can be defined from the line above, but it can
# also be defined here because setup.py has been exec'ed from
# setup_egg.py.
if not 'extra_setuptools_args' in globals():
    extra_setuptools_args = dict()


# Hard and soft dependency checking
package_check('numpy', INFO_VARS['NUMPY_MIN_VERSION'])
package_check('scipy', INFO_VARS['SCIPY_MIN_VERSION'])
package_check('nibabel', INFO_VARS['NIBABEL_MIN_VERSION'])
package_check('sympy', INFO_VARS['SYMPY_MIN_VERSION'])
def _mayavi_version(pkg_name):
    """Mayavi2 pruned enthought. namespace at 4.0.0
    """
    v = ''
    try:
        from mayavi import version
        v = version.version
        if v == '':
            v = '4.0.0' # must be the one in Debian
    except ImportError:
        from enthought.mayavi import version
        v = version.version
    return v
package_check('mayavi',
              INFO_VARS['MAYAVI_MIN_VERSION'],
              optional=True,
              version_getter=_mayavi_version)
# Cython can be a build dependency
def _cython_version(pkg_name):
    from Cython.Compiler.Version import version
    return version
package_check('cython',
              INFO_VARS['CYTHON_MIN_VERSION'],
              optional=True,
              version_getter=_cython_version,
              messages={'opt suffix': ' - you will not be able '
                        'to rebuild Cython source files into C files',
                        'missing opt': 'Missing optional build-time '
                        'package "%s"'}
              )

################################################################################
# commands for installing the data
from numpy.distutils.command.install_data import install_data
from numpy.distutils.command.build_ext import build_ext

def data_install_msgs():
    # Check whether we have data packages
    from nibabel.data import datasource_or_bomber
    DATA_PKGS = INFO_VARS['DATA_PKGS']
    templates = datasource_or_bomber(DATA_PKGS['nipy-templates'])
    example_data = datasource_or_bomber(DATA_PKGS['nipy-data'])
    for dpkg in (templates, example_data):
        if hasattr(dpkg, 'msg'): # a bomber object, warn
            log.warn('%s\n%s' % ('_'*80, dpkg.msg))


class MyInstallData(install_data):
    """ Subclass the install_data to generate data install warnings if necessary
    """
    def run(self):
        install_data.run(self)
        data_install_msgs()


class MyBuildExt(build_ext):
    """ Subclass the build_ext to generate data install warnings if
        necessary: warn at build == warn early
        This is also important to get a warning when run a 'develop'.
    """
    def run(self):
        build_ext.run(self)
        data_install_msgs()


cmdclass['install_data'] = MyInstallData
cmdclass['build_ext'] = MyBuildExt

################################################################################


def main(**extra_args):
    from numpy.distutils.core import setup

    setup(name=INFO_VARS['NAME'],
          maintainer=INFO_VARS['MAINTAINER'],
          maintainer_email=INFO_VARS['MAINTAINER_EMAIL'],
          description=INFO_VARS['DESCRIPTION'],
          long_description=INFO_VARS['LONG_DESCRIPTION'],
          url=INFO_VARS['URL'],
          download_url=INFO_VARS['DOWNLOAD_URL'],
          license=INFO_VARS['LICENSE'],
          classifiers=INFO_VARS['CLASSIFIERS'],
          author=INFO_VARS['AUTHOR'],
          author_email=INFO_VARS['AUTHOR_EMAIL'],
          platforms=INFO_VARS['PLATFORMS'],
          # version set by config.get_version() above
          requires=INFO_VARS['REQUIRES'],
          configuration = configuration,
          cmdclass = cmdclass,
          scripts = glob('scripts/*'),
          **extra_args)


if __name__ == "__main__":
    main(**extra_setuptools_args)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Test functions for models.regression
"""

import numpy as np

import scipy.linalg as spl

from ..regression import (OLSModel, ARModel, yule_walker, AREstimator,
                          ar_bias_corrector, ar_bias_correct)

from nose.tools import assert_equal, assert_true
from numpy.testing import assert_array_almost_equal, assert_array_equal


RNG = np.random.RandomState(20110902)
X = RNG.standard_normal((40,10))
Y = RNG.standard_normal((40,))


def test_OLS():
    model = OLSModel(design=X)
    results = model.fit(Y)
    assert_equal(results.df_resid, 30)


def test_AR():
    model = ARModel(design=X, rho=0.4)
    results = model.fit(Y)
    assert_equal(results.df_resid, 30)


def test_OLS_degenerate():
    Xd = X.copy()
    Xd[:,0] = Xd[:,1] + Xd[:,2]
    model = OLSModel(design=Xd)
    results = model.fit(Y)
    assert_equal(results.df_resid, 31)


def test_AR_degenerate():
    Xd = X.copy()
    Xd[:,0] = Xd[:,1] + Xd[:,2]
    model = ARModel(design=Xd, rho=0.9)
    results = model.fit(Y)
    assert_equal(results.df_resid, 31)


def test_yule_walker_R():
    # Test YW implementation against R results
    Y = np.array([1,3,4,5,8,9,10])
    N = len(Y)
    X = np.ones((N, 2))
    X[:,0] = np.arange(1,8)
    pX = spl.pinv(X)
    betas = np.dot(pX, Y)
    Yhat = Y - np.dot(X, betas)
    # R results obtained from:
    # >>> np.savetxt('yhat.csv', Yhat)
    # > yhat = read.table('yhat.csv')
    # > ar.yw(yhat$V1, aic=FALSE, order.max=2)
    def r_fudge(sigma, order):
        # Reverse fudge in ar.R calculation labeled as splus compatibility fix
        return sigma **2 * N / (N-order-1)
    rhos, sd = yule_walker(Yhat, 1, 'mle')
    assert_array_almost_equal(rhos, [-0.3004], 4)
    assert_array_almost_equal(r_fudge(sd, 1), 0.2534, 4)
    rhos, sd = yule_walker(Yhat, 2, 'mle')
    assert_array_almost_equal(rhos, [-0.5113, -0.7021], 4)
    assert_array_almost_equal(r_fudge(sd, 2), 0.1606, 4)
    rhos, sd = yule_walker(Yhat, 3, 'mle')
    assert_array_almost_equal(rhos, [-0.6737, -0.8204, -0.2313], 4)
    assert_array_almost_equal(r_fudge(sd, 3), 0.2027, 4)


def test_ar_estimator():
    # More or less a smoke test
    rng = np.random.RandomState(20110903)
    N = 100
    Y = rng.normal(size=(N,1)) * 10 + 100
    X = np.c_[np.linspace(-1,1,N), np.ones((N,))]
    my_model = OLSModel(X)
    results = my_model.fit(Y)
    are = AREstimator(my_model,2)
    rhos = are(results)
    assert_equal(rhos.shape, (2,))
    assert_true(np.all(np.abs(rhos <= 1)))
    rhos2 = ar_bias_correct(results, 2)
    assert_array_almost_equal(rhos, rhos2, 8)
    invM = ar_bias_corrector(my_model.design, my_model.calc_beta, 2)
    rhos3 = ar_bias_correct(results, 2, invM)
    assert_array_almost_equal(rhos2, rhos3)
    # Check orders 1 and 3
    rhos = ar_bias_correct(results, 1)
    assert_equal(rhos.shape, ())
    assert_true(abs(rhos) <= 1)
    rhos = ar_bias_correct(results, 3)
    assert_equal(rhos.shape, (3,))
    assert_true(np.all(np.abs(rhos) <= 1))
    # Make a 2D Y and try that
    Y = rng.normal(size=(N,12)) * 10 + 100
    results = my_model.fit(Y)
    rhos = are(results)
    assert_equal(rhos.shape, (2,12))
    assert_true(np.all(np.abs(rhos <= 1)))
    rhos2 = ar_bias_correct(results, 2)
    assert_array_almost_equal(rhos, rhos2, 8)
    rhos3 = ar_bias_correct(results, 2, invM)
    assert_array_almost_equal(rhos2, rhos3)
    # Passing in a simple array
    rhos4 = ar_bias_correct(results.resid, 2, invM)
    assert_array_almost_equal(rhos3, rhos4)
    # Check orders 1 and 3
    rhos = ar_bias_correct(results, 1)
    assert_equal(rhos.shape, (12,))
    assert_true(np.all(np.abs(rhos) <= 1))
    rhos = ar_bias_correct(results, 3)
    assert_equal(rhos.shape, (3,12))
    assert_true(np.all(np.abs(rhos) <= 1))
    # Try reshaping to 3D
    results.resid = results.resid.reshape((N,3,4))
    rhos = ar_bias_correct(results, 2)
    assert_array_almost_equal(rhos, rhos2.reshape((2,3,4)))



from scipy.optimize import (fmin as fmin_simplex,
                            fmin_powell,
                            fmin_cg,
                            fmin_bfgs,
                            fmin_ncg)
from ..optimize import fmin_steepest


def subdict(dic, keys):
    sdic = {}
    for k in keys:
        sdic[k] = dic[k]
    return sdic


def configure_optimizer(optimizer, fprime=None, fhess=None, **kwargs):
    """
    Return the minimization function
    """
    args = []
    kwargs['fprime'] = fprime
    kwargs['fhess'] = fhess
    kwargs['avextol'] = kwargs['xtol']

    if optimizer == 'simplex':
        keys = ('xtol', 'ftol', 'maxiter', 'maxfun')
        fmin = fmin_simplex
    elif optimizer == 'powell':
        keys = ('xtol', 'ftol', 'maxiter', 'maxfun')
        fmin = fmin_powell
    elif optimizer == 'cg':
        keys = ('gtol', 'maxiter', 'fprime')
        fmin = fmin_cg
    elif optimizer == 'bfgs':
        keys = ('gtol', 'maxiter', 'fprime')
        fmin = fmin_bfgs
    elif optimizer == 'ncg':
        args = [fprime]
        keys = ('avextol', 'maxiter', 'fhess')
        fmin = fmin_ncg
    elif optimizer == 'steepest':
        keys = ('xtol', 'ftol', 'maxiter', 'fprime')
        fmin = fmin_steepest
    else:
        raise ValueError('unknown optimizer: %s' % optimizer)

    return fmin, args, subdict(kwargs, keys)


def use_derivatives(optimizer):
    if optimizer in ('simplex', 'powell'):
        return False
    else:
        return True



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
"""
Get two images from the web (one mask image and one spmT image) and put them in
the nipy user dir - usually therefore at ``~/.nipy/tests/data``.

Author : Bertrand Thirion, 2009
"""

import os
try:
    from urllib2 import urlopen # Python 2
except ImportError:
    from urllib.request import urlopen # Python 3
import tarfile

from nibabel.data import get_nipy_user_dir

NIPY_DIR = get_nipy_user_dir()
DATA_DIR = os.path.join(NIPY_DIR, 'tests', 'data')


def get_second_level_dataset():
    """ Lightweight dataset for multi-subject analysis
    """
    # define several paths
    url = 'ftp://ftp.cea.fr/pub/dsv/madic/download/nipy'
    mask_image = os.path.join(DATA_DIR, 'mask.nii.gz')
    input_image = os.path.join(DATA_DIR, 'spmT_0029.nii.gz')
    group_data = os.path.join(DATA_DIR, 'group_t_images.tar.gz')

    # if needed create DATA_DIR
    if not os.path.exists(DATA_DIR):
        os.makedirs(DATA_DIR)
        assert os.path.exists(DATA_DIR)

    # download mask_image if necessary
    if not os.path.exists(mask_image):
        filename = 'mask.nii.gz'
        datafile = os.path.join(url, filename)
        fp = urlopen(datafile)
        local_file = open(mask_image, 'wb')
        local_file.write(fp.read())
        local_file.flush()
        local_file.close()

    # download input_image if necessary
    if not os.path.exists(input_image):
        filename = 'spmT_0029.nii.gz'
        datafile = os.path.join(url, filename)
        fp = urlopen(datafile)
        local_file = open(input_image, 'wb')
        local_file.write(fp.read())
        local_file.flush()
        local_file.close()

    # download group_data if necessary
    if not os.path.exists(group_data):
        filename = 'group_t_images.tar.gz'
        datafile = os.path.join(url, filename)
        fp = urlopen(datafile)
        local_file = open(group_data, 'wb')
        local_file.write(fp.read())
        local_file.flush()
        local_file.close()

    # untargzip group_data
    tar = tarfile.open(group_data)
    tar.extractall(DATA_DIR)
    tar.close()
    os.remove(group_data)
    return DATA_DIR


def get_first_level_dataset():
    """ Heavier dataset (30 MO) for first-level analysis
    """
    # define several paths
    url = 'ftp://ftp.cea.fr/pub/dsv/madic/download/nipy'
    raw_fmri = os.path.join(DATA_DIR, 's12069_swaloc1_corr.nii.gz')
    paradigm = os.path.join(DATA_DIR, 'localizer_paradigm.csv')

    # create DATA_DIR
    if not os.path.exists(DATA_DIR):
        os.makedirs(DATA_DIR)
        assert os.path.exists(DATA_DIR)

    # download mask_image if necessary
    if not os.path.exists(paradigm):
        print('Downloading mask image, this may take time')
        datafile = os.path.join(url, 'localizer_paradigm.csv')
        fp = urlopen(datafile)
        local_file = open(paradigm, 'wb')
        local_file.write(fp.read())
        local_file.flush()
        local_file.close()

    # download raw_fmri if necessary
    if not os.path.exists(raw_fmri):
        print('Downloading fmri image, this may take time')
        filename = 's12069_swaloc1_corr.nii.gz'
        datafile = os.path.join(url, filename)
        fp = urlopen(datafile)
        local_file = open(raw_fmri, 'wb')
        local_file.write(fp.read())
        local_file.flush()
        local_file.close()

    return DATA_DIR


if __name__ == '__main__':
    get_second_level_dataset()



import numpy as np

from ..utils import inique, permuted_signs, permuted_with_signs

from ..taitbryan import euler2mat

# Regular points around a sphere
_r13 = np.sqrt(1/3.0)
_r12 = np.sqrt(0.5)
sphere_points = (
        tuple(inique(permuted_with_signs([1, 0, 0]))) + 
        tuple(inique(permuted_with_signs([_r12, _r12, 0]))) + 
        tuple(inique(permuted_signs([_r13, _r13, _r13])))
    )

# Example rotations '''
euler_tuples = []
params = np.arange(-np.pi,np.pi,np.pi/2)
euler_tuples = tuple((x, y, z)
                     for x in params
                     for y in params
                     for z in params)
    
euler_mats = tuple(euler2mat(*t) for t in euler_tuples)




""" Testing ``isestimable`` in regression module
"""

import numpy as np

from ..regression import isestimable

from numpy.testing import (assert_almost_equal,
                           assert_array_equal)

from nose.tools import (assert_true, assert_false, assert_raises,
                        assert_equal, assert_not_equal)


def test_estimable():
    rng = np.random.RandomState(20120713)
    N, P = (40, 10)
    X = rng.normal(size=(N, P))
    C = rng.normal(size=(1, P))
    assert_true(isestimable(C, X))
    assert_true(isestimable(np.eye(P), X))
    for row in np.eye(P):
        assert_true(isestimable(row, X))
    X = np.ones((40, 2))
    assert_true(isestimable([1, 1], X))
    assert_false(isestimable([1, 0], X))
    assert_false(isestimable([0, 1], X))
    assert_false(isestimable(np.eye(2), X))
    halfX = rng.normal(size=(N, 5))
    X = np.hstack([halfX, halfX])
    assert_false(isestimable(np.hstack([np.eye(5), np.zeros((5, 5))]), X))
    assert_false(isestimable(np.hstack([np.zeros((5, 5)), np.eye(5)]), X))
    assert_true(isestimable(np.hstack([np.eye(5), np.eye(5)]), X))
    # Test array-like for design
    XL = X.tolist()
    assert_true(isestimable(np.hstack([np.eye(5), np.eye(5)]), XL))
    # Test ValueError for incorrect number of columns
    X = rng.normal(size=(N, 5))
    for n in range(1, 4):
        assert_raises(ValueError, isestimable, np.ones((n,)), X)
    assert_raises(ValueError, isestimable, np.eye(4), X)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Some simple examples and utility functions for resampling.
"""

import copy

import numpy as np

from scipy.ndimage import affine_transform

from nibabel.affines import from_matvec, to_matvec

from .interpolation import ImageInterpolator
from ..core.api import (Image, CoordinateMap, AffineTransform,
                        ArrayCoordMap, compose)

def resample_img2img(source, target, order=3):
    """  Resample `source` image to space of `target` image

    This wraps the resample function to resample one image onto another.
    The output of the function will give an image with shape of the
    target and data from the source

    Parameters
    ----------
    source : ``Image``
       Image instance that is to be resampled
    target : ``Image``
       Image instance to which source is resampled The output image will
       have the same shape as the target, and the same coordmap
    order : ``int``, optional
       What order of interpolation to use in `scipy.ndimage`

    Returns
    -------
    output : ``Image``
       Image with interpolated data and output.coordmap == target.coordmap

    Examples
    --------
    >>> from nipy.testing import funcfile, anatfile
    >>> from nipy.io.api import load_image
    >>> aimg_source = load_image(anatfile)
    >>> aimg_target = aimg_source
    >>> # in this case, we resample aimg to itself
    >>> resimg = resample_img2img(aimg_source, aimg_target)
    """
    sip, sop = source.coordmap.ndims
    tip, top = target.coordmap.ndims
    #print sip, sop, tip, top
    if sop != top:
        raise ValueError("source coordmap output dimension not equal "
                         "to target coordmap output dimension")
    mapping = np.eye(sop+1) # this would usually be 3+1
    resimg = resample(source, target.coordmap, mapping, target.shape, order=order)
    return resimg


def resample(image, target, mapping, shape, order=3):
    """ Resample `image` to `target` CoordinateMap

    Use a "world-to-world" mapping `mapping` and spline interpolation of a 
    `order`.

    Here, "world-to-world" refers to the fact that mapping should be a
    callable that takes a physical coordinate in "target" and gives a
    physical coordinate in "image".

    Parameters
    ----------
    image : Image instance
       image that is to be resampled
    target : CoordinateMap
       coordinate map for output image
    mapping : callable or tuple or array
       transformation from target.function_range to
       image.coordmap.function_range, i.e. 'world-to-world mapping'. Can
       be specified in three ways: a callable, a tuple (A, b)
       representing the mapping y=dot(A,x)+b or a representation of this
       mapping as an affine array, in homogeneous coordinates.
    shape : sequence of int
       shape of output array, in target.function_domain
    order : int, optional
       what order of interpolation to use in `scipy.ndimage`

    Returns
    -------
    output : Image instance
       with interpolated data and output.coordmap == target
    """
    if not callable(mapping):
        if type(mapping) is type(()):
            mapping = from_matvec(*mapping)
        # image world to target world mapping
        TW2IW = AffineTransform(target.function_range,
                                image.coordmap.function_range,
                                mapping)
    else:
        if isinstance(mapping, AffineTransform):
            TW2IW = mapping
        else:
            TW2IW = CoordinateMap(target.function_range,
                                  image.coordmap.function_range,
                                  mapping)
    # target voxel to image world mapping
    TV2IW = compose(TW2IW, target)
    # CoordinateMap describing mapping from target voxel to
    # image world coordinates
    if not isinstance(TV2IW, AffineTransform):
        # interpolator evaluates image at values image.coordmap.function_range,
        # i.e. physical coordinates rather than voxel coordinates
        grid = ArrayCoordMap.from_shape(TV2IW, shape)
        interp = ImageInterpolator(image, order=order)
        idata = interp.evaluate(grid.transposed_values)
        del(interp)
    else: # it is an affine transform, but, what if we compose?
        TV2IV = compose(image.coordmap.inverse(), TV2IW)
        if isinstance(TV2IV, AffineTransform): # still affine
            A, b = to_matvec(TV2IV.affine)
            idata = affine_transform(image.get_data(), A,
                                     offset=b,
                                     output_shape=shape,
                                     order=order)
        else: # not affine anymore
            interp = ImageInterpolator(image, order=order)
            grid = ArrayCoordMap.from_shape(TV2IV, shape)
            idata = interp.evaluate(grid.values)
            del(interp)
    return Image(idata, copy.copy(target))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This module is meant to reproduce the GLM analysis of fmristat.

    Liao et al. (2002).
TODO fix reference here

"""
__docformat__ = 'restructuredtext'

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This sub-package contains functions for clustering.
It might be removed in the future, and replaced 
by an optional dependence on scikit learn. 
"""

from nipy.testing import Tester

test = Tester().test
bench = Tester().bench 



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The Image class provides the interface which should be used
by users at the application level. The image provides a coordinate map,
and the data itself.

"""
__docformat__ = 'restructuredtext'

# You'd usually use nipy.core.api for these
from . import image
from .image import Image

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The main routine of this package that aims at performing the
extraction of ROIs from multisubject dataset using the localization
and activation strength of extracted regions.

This has been published in:
- Thirion et al. High level group analysis of FMRI data based on
Dirichlet process mixture models, IPMI 2007
- Thirion et al.
Accurate Definition of Brain Regions Position Through the
Functional Landmark Approach, MICCAI 2010

Author : Bertrand Thirion, 2006-2011
"""

import numpy as np
import scipy.stats as st

from .structural_bfls import build_LR
from nipy.algorithms.graph import wgraph_from_coo_matrix
from ...algorithms.statistics.empirical_pvalue import \
    NormalEmpiricalNull, three_classes_GMM_fit, gamma_gaussian_fit
from .hroi import HROI_as_discrete_domain_blobs

####################################################################
# Ancillary functions
####################################################################


def _relabel_(label, nl=None):
    """ Simple utilisity to relabel a pre-existing label vector

    Parameters
    ----------
    label: array of shape(n)
    nl: array of shape(p), where p<= label.max(), optional
        if None, the output is -1*np.ones(n)

    Returns
    -------
    new_label : array of shape (n)
    """
    if label.max() + 1 < np.size(nl):
        raise ValueError('incompatible values for label of nl')
    new_label = - np.ones(np.shape(label))
    if nl is not None:
        aux = np.arange(label.max() + 1)
        aux[0: np.size(nl)] = nl
        new_label[label > - 1] = aux[label[label > - 1]]
    return new_label


def signal_to_pproba(test, learn=None, method='prior', alpha=0.01, verbose=0):
    """Convert a set of z-values to posterior probabilities of not being active

    Parameters
    ----------
    test: array pf shape(n_samples),
           data that is assessed
    learn: array pf shape(n_samples), optional
           data to learn a mixture model
    method: string, optional, to be chosen within
            ['gauss_mixture', 'emp_null', 'gam_gauss', 'prior']
    alpha: float in the [0,1], optional,
           parameter that yields the prior probability that a region is active
           should be chosen close to 0
    """
    if method == 'gauss_mixture':
        prior_strength = 100
        fixed_scale = True
        bfp = three_classes_GMM_fit(
            learn, test, alpha, prior_strength, verbose, fixed_scale)
        bf0 = bfp[:, 1]
    elif method == 'emp_null':
        enn = NormalEmpiricalNull(learn)
        enn.learn()
        bf0 = np.reshape(enn.fdr(test), np.size(test))
    elif method == 'gam_gauss':
        bfp = gamma_gaussian_fit(learn, test, verbose)
        bf0 = bfp[:, 1]
    elif method == 'prior':
        y0 = st.norm.pdf(test)
        shape_, scale_ = 3., 2.
        y1 = st.gamma.pdf(test, shape_, scale=scale_)
        bf0 = np.ravel((1 - alpha) * y0 / (alpha * y1 + (1 - alpha) * y0))
    else:
        raise ValueError('Unknown method')
    return bf0


def compute_individual_regions(domain, lbeta, smin=5, theta=3.0,
                               method='gauss_mixture', verbose=0, reshuffle=0,
                               criterion='size', assign_val='weighted_mean'):
    """ Compute the individual regions that are real activation candidates

    Parameters
    ----------
    dom : StructuredDomain instance,
          generic descriptor of the space domain
    lbeta: an array of shape (nbnodes, subjects)
           the multi-subject statistical maps
    smin: int, optional
          minimal size of the regions to validate them
    theta: float, optional
           first level threshold
    method: string, optional,
           method that is used to provide priori significance
           can be 'prior', 'gauss_mixture', 'gam_gauss' or 'emp_null'
    verbose: verbosity mode, optional
    reshuffle: bool, otpional,
               if nonzero, reshuffle the positions; this affects bf and gfc
    criterion: string, optional,
               'size' or 'volume', thresholdding criterion
    assign_val: string, optional,
                to  be chosen in 'weighted mean', 'mean', 'min', 'max'
                heuristic to assigna  blob-level signal

    Returns
    -------
    bf list of nipy.labs.spatial_models.hroi.Nroi instances
       representing individual ROIs
       let nr be the number of terminal regions across subjects
    gf0, array of shape (nr)
         the mixture-based prior probability
         that the terminal regions are false positives
    sub, array of shape (nr)
         the subject index associated with the terminal regions
    gfc, array of shape (nr, coord.shape[1])
         the coordinates of the of the terminal regions

    Fixme
    -----
    Should allow for subject specific domains
    """
    bf = []
    gfc = []
    gf0 = []
    sub = []
    n_subj = lbeta.shape[1]
    nvox = lbeta.shape[0]

    for s in range(n_subj):
        # description in terms of blobs
        beta = np.reshape(lbeta[:, s], (nvox, 1))
        nroi = HROI_as_discrete_domain_blobs(
            domain, beta, threshold=theta, smin=smin)
        
        if nroi is not None and nroi.k > 0:
            bfm = nroi.representative_feature('signal', 'weighted mean')
            bfm = bfm[[nroi.select_id(id) for id in nroi.get_leaves_id()]]
            # get the regions position
            if reshuffle:
                nroi.reduce_to_leaves()
                ## randomize the positions
                ## by taking any local maximum of the image
                temp = np.argsort(np.random.rand(nvox))[:nroi.k]
                bfc = domain.coord[temp]
            else:
                mean_pos = np.asarray(
                    [np.mean(coords, 0) for coords in nroi.get_coord()])
                nroi.set_roi_feature('position', mean_pos)
                bfc = mean_pos[[nroi.select_id(id)
                                for id in nroi.get_leaves_id()]]
            gfc.append(bfc)

            # compute the prior proba of being null
            learn = np.squeeze(beta[beta != 0])
            bf0 = signal_to_pproba(bfm, learn, method)
            gf0.append(bf0)
            sub.append(s * np.ones(np.size(bfm)))
            
        bf.append(nroi)
    return bf, gf0, sub, gfc


def dpmm(gfc, alpha, g0, g1, dof, prior_precision, gf1, sub, burnin,
         spatial_coords=None, nis=1000, co_clust=False, verbose=False):
    """Apply the dpmm analysis to compute clusters from regions coordinates
    """
    from nipy.algorithms.clustering.imm import MixedIMM

    dim = gfc.shape[1]
    migmm = MixedIMM(alpha, dim)
    migmm.set_priors(gfc)
    migmm.set_constant_densities(null_dens=g0, prior_dens=g1)
    migmm._prior_dof = dof
    migmm._prior_scale = np.diag(prior_precision[0] / dof)
    migmm._inv_prior_scale_ = [np.diag(dof * 1. / (prior_precision[0]))]
    migmm.sample(gfc, null_class_proba=1 - gf1, niter=burnin, init=False,
                 kfold=sub)
    if verbose:
        print 'number of components: ', migmm.k

    # sampling
    if co_clust:
        like, pproba, co_clust = migmm.sample(
            gfc, null_class_proba=1 - gf1, niter=nis,
            sampling_points=spatial_coords, kfold=sub, co_clustering=co_clust)
        if verbose:
            print 'number of components: ', migmm.k

        return like, 1 - pproba, co_clust
    else:
        like, pproba = migmm.sample(
            gfc, null_class_proba=1 - gf1, niter=nis,
            sampling_points=spatial_coords, kfold=sub, co_clustering=co_clust)
    if verbose:
        print 'number of components: ', migmm.k

    return like, 1 - pproba


def bsa_dpmm(bf, gf0, sub, gfc, dmax, thq, ths, verbose=0):
    """ Estimation of the population level model of activation density using
    dpmm and inference

    Parameters
    ----------
    bf list of nipy.labs.spatial_models.hroi.HierarchicalROI instances
       representing individual ROIs
       let nr be the number of terminal regions across subjects
    gf0, array of shape (nr)
         the mixture-based prior probability
         that the terminal regions are true positives
    sub, array of shape (nr)
         the subject index associated with the terminal regions
    gfc, array of shape (nr, coord.shape[1])
         the coordinates of the of the terminal regions
    dmax float>0:
         expected cluster std in the common space in units of coord
    thq = 0.5 (float in the [0,1] interval)
        p-value of the prevalence test
    ths=0, float in the rannge [0,nsubj]
        null hypothesis on region prevalence that is rejected during inference
    verbose=0, verbosity mode

    Returns
    -------
    crmap: array of shape (nnodes):
           the resulting group-level labelling of the space
    LR: a instance of sbf.LandmarkRegions that describes the ROIs found
        in inter-subject inference
        If no such thing can be defined LR is set to None
    bf: List of  nipy.labs.spatial_models.hroi.Nroi instances
        representing individual ROIs
    p: array of shape (nnodes):
       likelihood of the data under H1 over some sampling grid
    """
    from nipy.algorithms.graph.field import field_from_coo_matrix_and_data
    dom = bf[0].domain
    n_subj = len(bf)

    crmap = - np.ones(dom.size, np.int)
    LR = None
    p = np.zeros(dom.size)
    if len(sub) < 1:
        return crmap, LR, bf, p

    sub = np.concatenate(sub).astype(np.int)
    gfc = np.concatenate(gfc)
    gf0 = np.concatenate(gf0)

    g0 = 1. / dom.local_volume.sum()

    # prepare the DPMM
    dim = dom.em_dim
    g1 = g0
    prior_precision = 1. / (dmax * dmax) * np.ones((1, dim))
    dof = 10
    burnin = 100
    nis = 1000
    # nis = number of iterations to estimate p
    p, q = dpmm(gfc, 0.5, g0, g1, dof, prior_precision, 1 - gf0,
               sub, burnin, dom.coord, nis)

    if verbose:
        h1, c1 = np.histogram((1 - gf0), bins=100)
        h2, c2 = np.histogram(q, bins=100)
        try:
            import matplotlib.pylab as pl
            pl.figure()
            pl.plot(1 - gf0, q, '.')
            pl.figure()
            pl.bar(c1[:len(h1)], h1, width=0.005)
            pl.bar(c2[:len(h2)] + 0.003, h2, width=0.005, color='r')
        except ImportError:
            pass
        print 'Number of candidate regions %i, regions found %i' % (
            np.size(q), q.sum())

    Fbeta = field_from_coo_matrix_and_data(dom.topology, p)
    _, label = Fbeta.custom_watershed(0, g0)

    # append some information to the hroi in each subject
    for s in range(n_subj):
        bfs = bf[s]
        if bfs.k > 0:
            leaves_pos = [bfs.select_id(k) for k in bfs.get_leaves_id()]
            us = - np.ones(bfs.k).astype(np.int)

            # set posterior proba
            lq = np.zeros(bfs.k)
            lq[leaves_pos] = q[sub == s]
            bfs.set_roi_feature('posterior_proba', lq)

            # set prior proba
            lq = np.zeros(bfs.k)
            lq[leaves_pos] = 1 - gf0[sub == s]
            bfs.set_roi_feature('prior_proba', lq)

            pos = np.asarray(
                [np.mean(coords, 0) for coords in bfs.get_coord()])
            midx = [np.argmin(np.sum((dom.coord - pos[k]) ** 2, 1))
                    for k in range(bfs.k)]
            j = label[np.array(midx)]
            us[leaves_pos] = j[leaves_pos]
            
            # when parent regions has similarly labelled children,
            # include it also
            us = bfs.make_forest().propagate_upward(us)
            bfs.set_roi_feature('label', us)

    # derive the group-level landmarks
    # with a threshold on the number of subjects
    # that are represented in each one
    LR, nl = build_LR(bf, thq, ths, dmax, verbose=verbose)

    # make a group-level map of the landmark position
    crmap = _relabel_(label, nl)
    return crmap, LR, bf, p


def bsa_dpmm2(bf, gf0, sub, gfc, dmax, thq, ths, verbose):
    """ Estimation of the population level model of activation density using
    dpmm and inference

    Parameters
    ----------
    bf list of nipy.labs.spatial_models.hroi.HierarchicalROI instances
       representing individual ROIs
       let nr be the number of terminal regions across subjects
    gf0, array of shape (nr)
         the mixture-based prior probability
         that the terminal regions are false positives
    sub, array of shape (nr)
         the subject index associated with the terminal regions
    gfc, array of shape (nr, coord.shape[1])
         the coordinates of the of the terminal regions
    dmax float>0:
         expected cluster std in the common space in units of coord
    thq = 0.5 (float in the [0,1] interval)
        p-value of the prevalence test
    ths=0, float in the rannge [0,nsubj]
        null hypothesis on region prevalence that is rejected during inference
    verbose=0, verbosity mode

    Returns
    -------
    crmap: array of shape (nnodes):
           the resulting group-level labelling of the space
    LR: a instance of sbf.LandmarkRegions that describes the ROIs found
        in inter-subject inference
        If no such thing can be defined LR is set to None
    bf: List of  nipy.labs.spatial_models.hroi.Nroi instances
        representing individual ROIs
    Coclust: array of shape (nr,nr):
             co-labelling matrix that gives for each pair of inputs
             how likely they are in the same class according to the model
    """
    dom = bf[0].domain
    n_subj = len(bf)
    crmap = - np.ones(dom.size, np.int)
    LR = None
    p = np.zeros(dom.size)
    if len(sub) < 1:
        return crmap, LR, bf, p

    sub = np.concatenate(sub).astype(np.int)
    gfc = np.concatenate(gfc)
    gf0 = np.concatenate(gf0)

    # prepare the DPMM
    g0 = 1. / (np.sum(dom.local_volume))
    g1 = g0
    prior_precision = 1. / (dmax * dmax) * np.ones((1, dom.em_dim), np.float)
    dof = 10
    burnin = 100
    nis = 300

    q, p, CoClust = dpmm(gfc, .5, g0, g1, dof, prior_precision, 1 - gf0,
                         sub, burnin, nis=nis, co_clust=True)

    cg = wgraph_from_coo_matrix(CoClust)
    if cg.E > 0:
        cg.remove_edges(cg.weights > .5)
    u = cg.cc()
    u[p < g0] = u.max() + 1 + np.arange(np.sum(p < g0))

    if verbose:
        cg.show(gfc)

    # append some information to the hroi in each subject
    for s in range(n_subj):
        bfs = bf[s]
        if bfs is not None:
            leaves = np.asarray(
                [bfs.select_id(id) for id in bfs.get_leaves_id()])
            us = - np.ones(bfs.k).astype(np.int)
            lq = np.zeros(bfs.k)
            lq[leaves] = q[sub == s]
            bfs.set_roi_feature('posterior_proba', lq)
            lq = np.zeros(bfs.k)
            lq[leaves] = 1 - gf0[sub == s]
            bfs.set_roi_feature('prior_proba', lq)
            us[leaves] = u[sub == s]

            # when parent regions has similarly labelled children,
            # include it also
            us = bfs.make_forest().propagate_upward(us)
            bfs.set_roi_feature('label', us)

    # derive the group-level landmarks
    # with a threshold on the number of subjects
    # that are represented in each one
    LR, nl = build_LR(bf, thq, ths, dmax, verbose=verbose)

    # make a group-level map of the landmark position
    crmap = - np.ones(dom.size)
    # not implemented at the moment

    return crmap, LR, bf, CoClust

###########################################################################
# Main functions
###########################################################################


def compute_BSA_simple(dom, lbeta, dmax, thq=0.5, smin=5, ths=0, theta=3.0,
                    method='prior', verbose=0):
    """ Compute the  Bayesian Structural Activation paterns
    simplified version

    Parameters
    ----------
    dom : StructuredDomain instance,
          Description of the spatial context of the data
    lbeta: an array of shape (nbnodes, subjects):
           the multi-subject statistical maps
    dmax float>0:
         expected cluster std in the common space in units of coord
    thq = 0.5 (float):
        posterior significance threshold
        should be in the [0,1] interval
    smin = 5 (int): minimal size of the regions to validate them
    theta = 3.0 (float): first level threshold
    method: string, optional,
            the method used to assess the prior significance of the regions
    verbose=0: verbosity mode

    Returns
    -------
    crmap: array of shape (nnodes):
           the resulting group-level labelling of the space
    LR: a instance of sbf.LandmarkRegions that describes the ROIs found
        in inter-subject inference
        If no such thing can be defined LR is set to None
    bf: List of  nipy.labs.spatial_models.hroi.Nroi instances
        representing individual ROIs
    p: array of shape (nnodes):
       likelihood of the data under H1 over some sampling grid

    Notes
    -----
    In that case, the DPMM is used to derive a spatial density of
    significant local maxima in the volume. Each terminal (leaf)
    region which is a posteriori significant enough is assigned to the
    nearest mode of this distribution

    fixme
    -----
    The number of itertions should become a parameter
    """
    bf, gf0, sub, gfc = compute_individual_regions(
        dom, lbeta, smin, theta, 'prior', verbose)

    crmap, LR, bf, p = bsa_dpmm(bf, gf0, sub, gfc, dmax, thq, ths, verbose)
    return crmap, LR, bf, p


def compute_BSA_quick(dom, lbeta, dmax, thq=0.5, smin=5, ths=0, theta=3.0,
                      verbose=0):
    """Idem compute_BSA_simple, but this one does not estimate the full density
    (on small datasets, it can be much faster)

    Parameters
    ----------
    dom : StructuredDomain instance,
          Description of the spatial context of the data
    lbeta: an array of shape (nbnodes, subjects):
           the multi-subject statistical maps
    dmax float>0:
         expected cluster std in the common space in units of coord
    thq = 0.5 (float):
        posterior significance threshold
        should be in the [0,1] interval
    smin = 5 (int): minimal size of the regions to validate them
    theta = 3.0 (float): first level threshold
    method: string, optional,
            the method used to assess the prior significance of the regions
    verbose=0: verbosity mode

    Returns
    -------
    crmap: array of shape (nnodes):
           the resulting group-level labelling of the space
    LR: a instance of sbf.LandmarkRegions that describes the ROIs found
        in inter-subject inference
        If no such thing can be defined LR is set to None
    bf: List of  nipy.labs.spatial_models.hroi.Nroi instances
        representing individual ROIs
    coclust: array of shape (nr, nr):
        co-labelling matrix that gives for each pair of cross_subject regions
        how likely they are in the same class according to the model
    """
    bf, gf0, sub, gfc = compute_individual_regions(
        dom, lbeta, smin, theta, 'prior', verbose)
    crmap, LR, bf, co_clust = bsa_dpmm2(
        bf, gf0, sub, gfc, dmax, thq, ths, verbose)
    return crmap, LR, bf, co_clust


def compute_BSA_loo(dom, lbeta, dmax, thq=0.5, smin=5, ths=0, theta=3.0,
                    verbose=0):
    """ Compute the  Bayesian Structural Activation paterns -
    with statistical validation

    Parameters
    ----------
    dom: StructuredDomain instance,
         Description of the spatial context of the data
    lbeta: an array of shape (nbnodes, subjects):
           the multi-subject statistical maps
    dmax float>0:
         expected cluster std in the common space in units of coord
    thq = 0.5 (float):
        posterior significance threshold
        should be in the [0,1] interval
    smin = 5 (int): minimal size of the regions to validate them
    theta = 3.0 (float): first level threshold
    method: string, optional,
            the method used to assess the prior significance of the regions
    verbose=0: verbosity mode

    Returns
    -------
    mll : float
        the average cross-validated log-likelihood across subjects
    ml0 : float
        the log-likelihood of the model under a global null hypothesis
    """
    n_subj = lbeta.shape[1]
    nvox = dom.size
    bf, gf0, sub, gfc = compute_individual_regions(
        dom, lbeta, smin, theta, 'gauss_mixture', verbose)

    p = np.zeros(nvox)
    g0 = 1. / (np.sum(dom.local_volume))
    if len(sub) < 1:
        return np.log(g0), np.log(g0)

    sub = np.concatenate(sub).astype(np.int)
    gfc = np.concatenate(gfc)
    gf0 = np.concatenate(gf0)

    # prepare the DPMM
    g1 = g0
    dim = dom.em_dim
    prior_precision = 1. / (dmax * dmax) * np.ones((1, dim), np.float)
    dof = 10
    burnin = 100
    nis = 300
    ll0 = []
    ll2 = []

    for s in range(n_subj):
        if np.sum(sub == s) > 0:
            spatial_coords = gfc[sub == s]
            p, q = dpmm(
                gfc[sub != s], 0.5, g0, g1, dof, prior_precision,
                1 - gf0[sub != s], sub[sub != s], burnin, spatial_coords, nis)
            pp = gf0[sub == s] * g0 + p * (1 - gf0[sub == s])
            ll2.append(np.mean(np.log(pp)))
            ll0.append(np.mean(np.log(g0)))

    ml0 = np.mean(np.array(ll0))
    mll = np.mean(np.array(ll2))
    if verbose:
        print 'average cross-validated log likelihood'
        print 'null model: ', ml0, ' alternative model: ', mll

    return mll, ml0



""" Chain transforms """

from .affine import Affine


class ChainTransform(object):
    def __init__(self, optimizable, pre=None, post=None):
        """ Create chain transform instance

        Parameters
        ----------
        optimizable : array or Transform
            Transform that we are optimizing.  If this is an array, then assume
            it's an affine matrix.
        pre : None or array or Transform, optional
            If not None, a transform that should be applied to points before
            applying the `optimizable` transform.  If an array, then assume it's
            an affine matrix.
        post : None or Transform, optional
            If not None, a transform that should be applied to points after
            applying any `pre` transform, and then the `optimizable`
            transform.  If an array, assume it's an affine matrix
        """
        if not hasattr(optimizable, 'param'):
            raise ValueError('Input transform should be optimizable')
        if not hasattr(optimizable, 'apply'):
            optimizable = Affine(optimizable)
        if not hasattr(pre, 'apply'):
            pre = Affine(pre)
        if not hasattr(post, 'apply'):
            post = Affine(post)
        self.optimizable = optimizable
        self.pre = pre
        self.post = post

    def apply(self, pts):
        """ Apply full transformation to points `pts`

        If there are N points, then `pts` will be N by 3

        Parameters
        ----------
        pts : array-like
            array of points

        Returns
        -------
        transformed_pts : array
            N by 3 array of transformed points
        """
        composed = self.post.compose(self.optimizable.compose(self.pre))
        return composed.apply(pts)

    def _set_param(self, param):
        self.optimizable.param = param
    def _get_param(self):
        return self.optimizable.param
    param = property(_get_param, _set_param, None, 'get/set param')



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Pseudo-package for some important statistics symbols

For example:

>>> from nipy.algorithms.statistics.api import Formula
"""
from .formula import formulae
from .formula.formulae import (Formula, Factor, Term, terms, make_recarray,
                               natural_spline)
from .models import (model, regression, glm, family)
from .models.regression import (OLSModel, ARModel, WLSModel, isestimable)




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Test functions for formulae
"""

import numpy as np
import sympy

from nipy.fixes.sympy.utilities.lambdify import implemented_function

from .. import formulae as F
from ..formulae import terms, Term

from nibabel.py3k import asbytes

from nose.tools import (assert_true, assert_equal, assert_false,
                        assert_raises)

from numpy.testing import assert_almost_equal, assert_array_equal

def test_terms():
    t = terms('a')
    assert_true(isinstance(t, Term))
    a, b, c = Term('a'), Term('b'), Term('c')
    assert_equal(t, a)
    ts = terms(('a', 'b', 'c'))
    assert_equal(ts, (a, b, c))
    # a string without separator chars returns one symbol.  This is the
    # sympy 0.7 behavior
    assert_equal(terms('abc'), Term('abc'))
    # separators return multiple symbols
    assert_equal(terms('a b c'), (a, b, c))
    assert_equal(terms('a, b, c'), (a, b, c))
    # no arg is an error
    assert_raises(TypeError, terms)
    # but empty arg returns empty tuple
    assert_equal(terms(()), ())
    # Test behavior of deprecated each_char kwarg
    try:
        res = terms('abc', each_char=False)
    except TypeError:
        return
    assert_equal(res, Term('abc'))
    assert_equal(terms('abc', each_char=True), (a, b, c))


def test_getparams_terms():
    t = F.Term('t')
    x, y, z = [sympy.Symbol(l) for l in 'xyz']
    yield assert_equal, set(F.getparams(x*y*t)), set([x,y])
    yield assert_equal, set(F.getterms(x*y*t)), set([t])

    matrix_expr = np.array([[x,y*t],[y,z]])
    yield assert_equal, set(F.getparams(matrix_expr)), set([x,y,z])
    yield assert_equal, set(F.getterms(matrix_expr)), set([t])


def test_formula_params():
    t = F.Term('t')
    x, y = [sympy.Symbol(l) for l in 'xy']
    f = F.Formula([t*x,y])
    yield assert_equal, set(f.params), set([x,y] + list(f.coefs.values()))


def test_contrast1():
    x = F.Term('x')
    yield assert_equal, x, x+x
    y = F.Term('y')
    z = F.Term('z')
    f = F.Formula([x,y])
    arr = F.make_recarray([[3,5,4],[8,21,-1],[4,6,-2]], 'xyz')
    D, C = f.design(arr, contrasts={'x':x.formula,
                                    'diff':F.Formula([x-y]),
                                    'sum':F.Formula([x+y]),
                                    'both':F.Formula([x-y,x+y])})
    yield assert_almost_equal, C['x'], np.array([1,0])
    yield assert_almost_equal, C['diff'], np.array([1,-1])
    yield assert_almost_equal, C['sum'], np.array([1,1])
    yield assert_almost_equal, C['both'], np.array([[1,-1],[1,1]])

    f = F.Formula([x,y,z])
    arr = F.make_recarray([[3,5,4],[8,21,-1],[4,6,-2]], 'xyz')
    D, C = f.design(arr, contrasts={'x':x.formula,
                                    'diff':F.Formula([x-y]),
                                    'sum':F.Formula([x+y]),
                                    'both':F.Formula([x-y,x+y])})
    yield assert_almost_equal, C['x'], np.array([1,0,0])
    yield assert_almost_equal, C['diff'], np.array([1,-1,0])
    yield assert_almost_equal, C['sum'], np.array([1,1,0])
    yield assert_almost_equal, C['both'], np.array([[1,-1,0],[1,1,0]])


def test_formula_from_recarray():
    D = np.rec.array([
            (43, 51, 30, 39, 61, 92, 'blue'),
            (63, 64, 51, 54, 63, 73, 'blue'),
            (71, 70, 68, 69, 76, 86, 'red'),
            (61, 63, 45, 47, 54, 84, 'red'),
            (81, 78, 56, 66, 71, 83, 'blue'),
            (43, 55, 49, 44, 54, 49, 'blue'),
            (58, 67, 42, 56, 66, 68, 'green'),
            (71, 75, 50, 55, 70, 66, 'green'),
            (72, 82, 72, 67, 71, 83, 'blue'),
            (67, 61, 45, 47, 62, 80, 'red'),
            (64, 53, 53, 58, 58, 67, 'blue'),
            (67, 60, 47, 39, 59, 74, 'green'),
            (69, 62, 57, 42, 55, 63, 'blue'),
            (68, 83, 83, 45, 59, 77, 'red'),
            (77, 77, 54, 72, 79, 77, 'red'),
            (81, 90, 50, 72, 60, 54, 'blue'),
            (74, 85, 64, 69, 79, 79, 'green'),
            (65, 60, 65, 75, 55, 80, 'green'),
            (65, 70, 46, 57, 75, 85, 'red'),
            (50, 58, 68, 54, 64, 78, 'red'),
            (50, 40, 33, 34, 43, 64, 'blue'),
            (64, 61, 52, 62, 66, 80, 'blue'),
            (53, 66, 52, 50, 63, 80, 'red'),
            (40, 37, 42, 58, 50, 57, 'red'),
            (63, 54, 42, 48, 66, 75, 'blue'),
            (66, 77, 66, 63, 88, 76, 'blue'),
            (78, 75, 58, 74, 80, 78, 'red'),
            (48, 57, 44, 45, 51, 83, 'blue'),
            (85, 85, 71, 71, 77, 74, 'red'),
            (82, 82, 39, 59, 64, 78, 'blue')], 
                     dtype=[('y', 'i8'),
                            ('x1', 'i8'),
                            ('x2', 'i8'),
                            ('x3', 'i8'),
                            ('x4', 'i8'),
                            ('x5', 'i8'),
                            ('x6', '|S5')])
    f = F.Formula.fromrec(D, drop='y')
    assert_equal(set([str(t) for t in f.terms]),
                 set(['x1', 'x2', 'x3', 'x4', 'x5',
                      'x6_green', 'x6_blue', 'x6_red']))
    assert_equal(set([str(t) for t in f.design_expr]),
                 set(['x1', 'x2', 'x3', 'x4', 'x5',
                      'x6_green', 'x6_blue', 'x6_red']))


def test_random_effects():
    subj = F.make_recarray([2,2,2,3,3], 's')
    subj_factor = F.Factor('s', [2,3])

    c = F.RandomEffects(subj_factor.terms, sigma=np.array([[4,1],[1,6]]))
    C = c.cov(subj)
    assert_almost_equal(C, [[4,4,4,1,1],
                            [4,4,4,1,1],
                            [4,4,4,1,1],
                            [1,1,1,6,6],
                            [1,1,1,6,6]])
    # Sympy 0.7.0 does not cancel 1.0 * A to A; however, the dot product in the
    # covariance calculation returns floats, which are them multiplied by the
    # terms to give term * 1.0, etc.  We just insert the annoying floating point
    # here for the test, relying on sympy to do the same thing here as in the
    # dot product
    a = sympy.Symbol('a') * 1.0
    b = sympy.Symbol('b') * 1.0
    c = F.RandomEffects(subj_factor.terms, sigma=np.array([[a,0],[0,b]]))
    C = c.cov(subj)
    t = np.equal(C, [[a,a,a,0,0],
                     [a,a,a,0,0],
                     [a,a,a,0,0],
                     [0,0,0,b,b],
                     [0,0,0,b,b]])
    assert_true(np.alltrue(t))


def test_design_expression():
    t1 = F.Term("x")
    t2 = F.Term('y')
    f = t1.formula + t2.formula
    assert_true(str(f.design_expr) in ['[x, y]', '[y, x]'])


def test_formula_property():
    # Check that you can create a Formula with one term
    t1 = F.Term("x")
    f = t1.formula
    assert_equal(f.design_expr, [t1])


def test_mul():
    f = F.Factor('t', [2,3])
    f2 = F.Factor('t', [2,3,4])
    t2 = f['t_2']
    x = F.Term('x')
    yield assert_equal, t2, t2*t2
    yield assert_equal, f, f*f
    yield assert_false, f == f2
    yield assert_equal, set((t2*x).atoms()), set([t2,x])


def test_make_recarray():
    m = F.make_recarray([[3,4],[4,6],[7,9]], 'wv', [np.float, np.int])
    assert_equal(m.dtype.names, ('w', 'v'))
    m2 = F.make_recarray(m, 'xy')
    assert_equal(m2.dtype.names, ('x', 'y'))


def test_str_formula():
    t1 = F.Term('x')
    t2 = F.Term('y')
    f = F.Formula([t1, t2])
    yield assert_equal, str(f), "Formula([x, y])"


def test_design():
    # Check that you get the design matrix we expect
    t1 = F.Term("x")
    t2 = F.Term('y')

    n = F.make_recarray([2,4,5], 'x')
    assert_almost_equal(t1.formula.design(n)['x'], n['x'])

    f = t1.formula + t2.formula
    n = F.make_recarray([(2,3),(4,5),(5,6)], 'xy')

    assert_almost_equal(f.design(n)['x'], n['x'])
    assert_almost_equal(f.design(n)['y'], n['y'])

    f = t1.formula + t2.formula + F.I + t1.formula * t2.formula
    assert_almost_equal(f.design(n)['x'], n['x'])
    assert_almost_equal(f.design(n)['y'], n['y'])
    assert_almost_equal(f.design(n)['1'], 1)
    assert_almost_equal(f.design(n)['x*y'], n['x']*n['y'])
    # drop x field, check that design raises error
    ny = np.recarray(n.shape, dtype=[('x', n.dtype['x'])])
    ny['x'] = n['x']
    assert_raises(ValueError, f.design, ny)
    n = np.array([(2,3,'a'),(4,5,'b'),(5,6,'a')], np.dtype([('x', np.float),
                                                            ('y', np.float),
                                                            ('f', 'S1')]))
    f = F.Factor('f', ['a','b'])
    ff = t1.formula * f + F.I
    assert_almost_equal(ff.design(n)['f_a*x'], n['x']*[1,0,1])
    assert_almost_equal(ff.design(n)['f_b*x'], n['x']*[0,1,0])
    assert_almost_equal(ff.design(n)['1'], 1)


def test_design_inputs():
    # Check we can send in fields of type 'S', 'U', 'O' for design
    regf = F.Formula(F.terms('x, y'))
    f = F.Factor('f', ['a', 'b'])
    ff = regf + f
    for field_type in ('S1', 'U1', 'O'):
        data = np.array([(2, 3, 'a'),
                         (4, 5, 'b'),
                         (5, 6, 'a')],
                        dtype = [('x', np.float),
                                 ('y', np.float),
                                 ('f', field_type)])
        assert_array_equal(ff.design(data, return_float=True),
                           [[2, 3, 1, 0],
                            [4, 5, 0, 1],
                            [5, 6, 1, 0]])


def test_formula_inputs():
    # Check we can send in fields of type 'S', 'U', 'O' for factor levels
    level_names = ['red', 'green', 'blue']
    for field_type in ('S', 'U', 'O'):
        levels = np.array(level_names, dtype=field_type)
        f = F.Factor('myname', levels)
        assert_equal(f.levels, level_names)
    # Sending in byte objects
    levels = [asbytes(L) for L in level_names]
    f = F.Factor('myname', levels)
    assert_equal(f.levels, level_names)


def test_alias():
    x = F.Term('x')
    f = implemented_function('f', lambda x: 2*x)
    g = implemented_function('g', lambda x: np.sqrt(x))
    ff = F.Formula([f(x), g(x)**2])
    n = F.make_recarray([2,4,5], 'x')
    assert_almost_equal(ff.design(n)['f(x)'], n['x']*2)
    assert_almost_equal(ff.design(n)['g(x)**2'], n['x'])


def test_factor_getterm():
    fac = F.Factor('f', 'ab')
    yield assert_equal, fac['f_a'], fac.get_term('a')
    fac = F.Factor('f', [1,2])
    yield assert_equal, fac['f_1'], fac.get_term(1)
    fac = F.Factor('f', [1,2])
    yield assert_raises, ValueError, fac.get_term, '1'
    m = fac.main_effect
    yield assert_equal, set(m.terms), set([fac['f_1']-fac['f_2']])


def test_stratify():
    fac = F.Factor('x', [2,3])

    y = sympy.Symbol('y')
    f = sympy.Function('f')
    yield assert_raises, ValueError, fac.stratify, f(y)


def test_nonlin1():
    # Fit an exponential curve, with the exponent stratified by a factor
    # with a common intercept and multiplicative factor in front of the
    # exponential
    x = F.Term('x')
    fac = F.Factor('f', 'ab')
    f = F.Formula([sympy.exp(fac.stratify(x).mean)]) + F.I
    params = F.getparams(f.mean)
    assert_equal(set([str(p) for p in params]),
                 set(['_x0', '_x1', '_b0', '_b1']))
    test1 = set(['1',
                 'exp(_x0*f_a + _x1*f_b)',
                 '_b0*f_a*exp(_x0*f_a + _x1*f_b)',
                 '_b0*f_b*exp(_x0*f_a + _x1*f_b)'])
    test2 = set(['1',
                 'exp(_x0*f_a + _x1*f_b)',
                 '_b1*f_a*exp(_x0*f_a + _x1*f_b)',
                 '_b1*f_b*exp(_x0*f_a + _x1*f_b)'])
    assert_true(test1 or test2)
    n = F.make_recarray([(2,3,'a'),(4,5,'b'),(5,6,'a')], 'xyf', ['d','d','S1'])
    p = F.make_recarray([1,2,3,4], ['_x0', '_x1', '_b0', '_b1'])
    A = f.design(n, p)
    print A, A.dtype


def test_intercept():
    dz = F.make_recarray([2,3,4],'z')
    v = F.I.design(dz, return_float=False)
    assert_equal(v.dtype.names, ('intercept',))


def test_nonlin2():
    dz = F.make_recarray([2,3,4],'z')
    z = F.Term('z')
    t = sympy.Symbol('th')
    p = F.make_recarray([3], ['tt'])
    f = F.Formula([sympy.exp(t*z)])
    yield assert_raises, ValueError, f.design, dz, p


def test_Rintercept():
    x = F.Term('x')
    y = F.Term('x')
    xf = x.formula
    yf = y.formula
    newf = (xf+F.I)*(yf+F.I)
    assert_equal(set(newf.terms), set([x,y,x*y,sympy.Number(1)]))


def test_return_float():
    x = F.Term('x')
    f = F.Formula([x,x**2])
    xx= F.make_recarray(np.linspace(0,10,11), 'x')
    dtype = f.design(xx).dtype
    yield assert_equal, set(dtype.names), set(['x', 'x**2'])
    dtype = f.design(xx, return_float=True).dtype
    yield assert_equal, dtype, np.float


def test_subtract():
    x, y, z = [F.Term(l) for l in 'xyz']
    f1 = F.Formula([x,y])
    f2 = F.Formula([x,y,z])
    f3 = f2 - f1
    yield assert_equal, set(f3.terms), set([z])
    f4 = F.Formula([y,z])
    f5 = f1 - f4
    yield assert_equal, set(f5.terms), set([x])


def test_subs():
    t1 = F.Term("x")
    t2 = F.Term('y')
    z = F.Term('z')
    f = F.Formula([t1, t2])
    g = f.subs(t1, z)
    yield assert_equal, list(g.terms), [z, t2]


def test_natural_spline():
    xt=F.Term('x')

    ns=F.natural_spline(xt, knots=[2,6,9])
    xx= F.make_recarray(np.linspace(0,10,101), 'x')
    dd=ns.design(xx, return_float=True)
    xx = xx['x']
    yield assert_almost_equal, dd[:,0], xx
    yield assert_almost_equal, dd[:,1], xx**2
    yield assert_almost_equal, dd[:,2], xx**3
    yield assert_almost_equal, dd[:,3], (xx-2)**3*np.greater_equal(xx,2)
    yield assert_almost_equal, dd[:,4], (xx-6)**3*np.greater_equal(xx,6)
    yield assert_almost_equal, dd[:,5], (xx-9)**3*np.greater_equal(xx,9)

    ns=F.natural_spline(xt, knots=[2,9,6], intercept=True)
    xx= F.make_recarray(np.linspace(0,10,101), 'x')
    dd=ns.design(xx, return_float=True)
    xx = xx['x']
    yield assert_almost_equal, dd[:,0], 1
    yield assert_almost_equal, dd[:,1], xx
    yield assert_almost_equal, dd[:,2], xx**2
    yield assert_almost_equal, dd[:,3], xx**3
    yield assert_almost_equal, dd[:,4], (xx-2)**3*np.greater_equal(xx,2)
    yield assert_almost_equal, dd[:,5], (xx-9)**3*np.greater_equal(xx,9)
    yield assert_almost_equal, dd[:,6], (xx-6)**3*np.greater_equal(xx,6)


def test_factor_term():
    # Test that byte strings, unicode strings and objects convert correctly
    for nt in 'S3', 'U3', 'O':
        ndt = np.dtype(nt)
        for lt in 'S3', 'U3', 'O':
            ldt = np.dtype(lt)
            name = np.asscalar(np.array('foo', ndt))
            level = np.asscalar(np.array('bar', ldt))
            ft = F.FactorTerm(name, level)
            assert_equal(str(ft), 'foo_bar')
